/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is favlib, the Windows-compatible bookmarks library.
 *
 * The Initial Developer of the Original Code is
 * Forest Wilkinson.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Forest Wilkinson
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


/*
favlib.cpp

Copyright 2004-2008 Forest Wilkinson

A library for reading and manipulating Windows favorites.


Implementor's Notes:

I started this project as a labor of love, hoping it would lead
to full Windows Favorites support in Mozilla Firefox.  Let's face
it: using Firefox bookmarks would be a lot more convenient if
they were well integrated with the desktop environment.

I first discussed this online in the Mozilla bug 56765 comments:
https://bugzilla.mozilla.org/show_bug.cgi?id=56765

There is a lot more code in here than I ever expected to write just
for simple web bookmark operations.  I'm not a fan of bloat, but
I'm even less a fan of windows-style APIs, and of software that
requires the absolute latest release of an operating system.
I therefore ended up writing lots of wrapper classes and helper
functions.  My intent was to make the code somewhat portable (among
windows versions), memory-friendly (using windows native allocators),
relatively safe when accessing undocumented data structures, and
generally more pleasant than win32 and COM APIs.  I would probably
do some things differently if I were to begin this project today,
but it works, so I don't think I'll be rewriting it.

This library uses:

- C++ templates.
- C++ exceptions.
- An anonymous C++ namespace.
  (It could be removed.  I only used it to avoid declaring everything static.)
- Windows API calls.
- Windows shell COM interfaces.
- Windows shell COM memory allocators.
- MS Active Template Library string conversion macros.

It does not use:

- C++ runtime type information.
- Any Windows call or interface that (according to the MS documentation)
  requires special service packs or add-on libraries.

It can be built with either Unicode or ANSI strings in the public
interface, via the windows TCHAR type.  It is intended to be
compatible with every version of windows that natively supports
the win95 or winxp shell UI.

I developed this library using Visual C++ 6.0, and tested with
Windows 2000.  I have no means of testing with other versions at the
time of this writing.

Perhaps the most difficult feature was supporting the user-modifiable
(drag & drop) display order that internet explorer uses for favorites.
The ordering information is stored in undocumented Windows registry
data structures, and although a few people on the net have posted
notes about these, none of them seemed complete or entirely accurate.
See the comment block later in this file for more details.


Known Issues: xxx
data is being saved, but already-running ie doesn't seem to re-read it.
remove() was seen to preserve our view of ordering, but reset ie to default.

TODO: xxx
add a countallchildren() function, to support favorites import progress bars?
add input range checking to every public function?
add a refresh() function, for use after change notification?
add winerror.h facility flags to win error codes?

*/


#include "favlib.h"

#include <atlbase.h> // for string conversion macros (see USES_CONVERSION)
#include <intshcut.h>
#include <io.h>
#include <shlobj.h>
#include <windows.h>

#include <algorithm>
#include <cassert>
#include <cstring>
#include <new>
#include <vector>

using namespace std;

namespace { // begin private namespace; this stuff is for internal use.


/////////////////////////////////////////////////////////////////////////////
// COM interface helpers


// Calls a COM interface's Release() method:
template <class T>
inline void com_release( T *ptr) throw()
    { if (ptr) ptr->Release(); }


// Calls a COM interface's AddRef() method:
template <class T>
inline void com_addref( T *ptr) throw()
    { if (ptr) ptr->AddRef(); }


// A class like std::auto_ptr, but for COM interfaces.
// Not for STL containers; copy & assignment steal ownership of internal data.
template <class T>
class auto_ifptr
    {
    public:
        explicit auto_ifptr( T *ptr = 0) throw()
            : m_ptr(ptr) {}
        auto_ifptr( auto_ifptr &other) throw()
            : m_ptr(other.disown()) {}
        auto_ifptr& operator=( auto_ifptr &other) throw()
            { reset( other.disown()); return *this; }
        ~auto_ifptr() throw()
            { com_release( m_ptr); }

        void reset( T *ptr = 0) throw()
            { if (m_ptr != ptr) { com_release( m_ptr); m_ptr = ptr; } }

        T *get() throw()
            { return m_ptr; }
        bool is_zero() const throw()
            { return m_ptr == 0; }

        T& operator*() throw() { return *get(); }
        T* operator->() throw() { return get(); }

        auto_ifptr copy() const throw()
            { com_addref( m_ptr); return auto_ifptr(m_ptr); }

    protected:
        T *disown() throw() // just like std::auto_ptr::release()
            { T *tmp = m_ptr; m_ptr = 0; return tmp; }

    private:
        T *m_ptr;
    };


// Returns a different interface to the given object, or throws an exception.
template <class OUTTYPE, class INTYPE>
auto_ifptr<OUTTYPE> query_iface(
    auto_ifptr<INTYPE> &in_iface, const IID &out_iface_id)
    {
    OUTTYPE *ptr;
    HRESULT result = in_iface->QueryInterface(
        out_iface_id, reinterpret_cast<void**>(&ptr));
    if (FAILED( result))
        {
        throw favlib_com_error( result);
        }
    return ptr;
    }


/////////////////////////////////////////////////////////////////////////////
// COM memory helpers


// We use these functions to access the Windows shell's memory allocator:
// Freeing or reallocating a 0 pointer is safe with these functions.
// Sizes are in bytes, no matter what pointer type is instantiated.

template <class T>
inline T *com_alloc( unsigned size) throw()
    { return reinterpret_cast<T *>(CoTaskMemAlloc( size)); }

template <class T>
inline T *com_realloc( T *mem, unsigned size) throw()
    { return reinterpret_cast<T *>(CoTaskMemRealloc( mem, size)); }

template <class T>
inline void com_free( T *mem) throw()
    { CoTaskMemFree( mem); }


/////////////////////////////////////////////////////////////////////////////
// COM memory pointer class


// A class like std::auto_ptr, but for memory allocated with COM allocator.
// Sizes are in bytes, no matter what pointer type is instantiated.
// Not for STL containers; copy & assignment steal ownership of internal data.
template <class T>
class auto_memptr
    {
    public:
        explicit auto_memptr( T *buf = 0) throw()
            : m_buf(buf) {}
        auto_memptr( auto_memptr &other) throw()
            : m_buf(other.disown()) {}
        auto_memptr& operator=( auto_memptr &other) throw()
            { reset( other.disown()); return *this; }
        ~auto_memptr() throw()
            { com_free( m_buf); }

        void reset( T *buf = 0) throw()
            { if (m_buf != buf) { com_free( m_buf); m_buf = buf; } }

        T *get() throw()
            { return m_buf; }
        const T *get() const throw()
            { return m_buf; }
        bool is_zero() const throw()
            { return get() == 0; }

        auto_memptr& resize( unsigned size);

        static auto_memptr<T> alloc( unsigned size); // throws on failure

    protected:
        T *disown() throw() // just like std::auto_ptr::release()
            { T *tmp = m_buf; m_buf = 0; return tmp; }

    private:
        T *m_buf;
    };


// Returns a new buffer of the given size (in bytes), or throws an exception.
template <class T>
auto_memptr<T> auto_memptr<T>::alloc( unsigned size)
    {
    T *tmp = com_alloc<T>( size);
    if (!tmp)
        {
        throw favlib_bad_alloc();
        }
    return tmp;
    }


// Acts like realloc(), but throws an exception (preserving m_buf) on failure.
template <class T>
auto_memptr<T>& auto_memptr<T>::resize( unsigned size)
    {
    T *tmp = com_realloc( m_buf, size);
    if (size && !tmp)
        {
        throw favlib_bad_alloc();
        }
    m_buf = tmp;
    return *this;
    }


/////////////////////////////////////////////////////////////////////////////
// An array class that uses the COM allocator.

/*
Rationale:

I chose to write this class instead of using a standard container class,
for these reasons:

It offers a pointer to raw memory, to which Windows API functions like
RegQueryValueEx() can write their output.

It does not impose such strict requirements on its contained objects
as STL containers do.

It can be used as a return value, without having to copy its contained
data.

It remains small, using only two words of RAM and performing no dynamic
allocations, until objects are actually stored.

It can resize() in place, without having to call any copy constructors
on already-stored objects.
*/


// Not for STL containers; copy & assignment steal ownership of internal data.
template <class T>
class auto_array
    {
    public:
        auto_array(): m_len(0) {}
        explicit auto_array( unsigned len);
        auto_array( unsigned len, const T &val); // init all elements to val
        auto_array( auto_array &o) // take ownership of the data
            : m_data(o.m_data), m_len(o.m_len) { o.m_len = 0; }
        auto_array& operator=( auto_array &o) // take ownership of the data
            { m_data = o.m_data; m_len = o.m_len;
            if (o.m_data.is_zero()) { o.m_len = 0; } return *this; }
        ~auto_array();

        void erase() { *this = auto_array<T>(); }
        void resize( unsigned len); // preserves contents

        // These are range checked: (index past end gets changed to end index.)
        T& move_item( unsigned to_index, unsigned from_index);
        T& insert_item( unsigned index) // returns reference to inserted item
            { resize( m_len + 1); return move_item( index, m_len - 1); }
        void erase_item( unsigned index)
            { move_item( m_len, index); resize( m_len - 1); }

        unsigned length() const throw() { return m_len; }

        // These are not range checked:
        T *data() throw()
            { return m_data.get(); }
        const T *data() const throw()
            { return m_data.get(); }
        T *begin() throw()
            { return m_data.get(); }
        const T *begin() const throw()
            { return m_data.get(); }
        T *end() throw()
            { return m_data.get() + m_len; }
        const T *end() const throw()
            { return m_data.get() + m_len; }
        T& operator[]( unsigned index) throw()
            { return m_data.get()[index]; }
        const T& operator[]( unsigned index) const throw()
            { return m_data.get()[index]; }

    private:
        auto_memptr<T> m_data;
        unsigned m_len;
    };


template <class T>
auto_array<T>::auto_array( unsigned len)
    : m_data(auto_memptr<T>::alloc(sizeof(T) * len)), m_len(len)
    {
    for (T *p = m_data.get(); p != m_data.get() + m_len; ++p)
        {
        new(p) T;
        }
    }


template <class T>
auto_array<T>::auto_array( unsigned len, const T &val)
    : m_data(auto_memptr<T>::alloc(sizeof(T) * len)), m_len(len)
    {
    for (T *p = m_data.get(); p != m_data.get() + m_len; ++p)
        {
        new(p) T(val);
        }
    }


template <class T>
auto_array<T>::~auto_array()
    {
    for (T *p = m_data.get(); p != m_data.get() + m_len; ++p)
        {
        p->~T();
        }
    }


template <class T>
void auto_array<T>::resize( unsigned len)
    {
    if (len < m_len)
        {
        for (T *p = begin() + len; p != begin() + m_len; ++p)
            {
            p->~T();
            }
        m_len = len;
        }
    m_data.resize( sizeof(T) * len);
    if (m_len < len)
        {
        for (T *p = begin() + m_len; p != begin() + len; ++p)
            {
            new(p) T;
            }
        m_len = len;
        }
    }


// Note: The moved item will be placed before the item at to_index.
template <class T>
T& auto_array<T>::move_item( unsigned to_index, unsigned from_index)
    {
    // Fix up indexes:
    if (to_index > from_index)
        {
        to_index -= 1;  // Insert just before the item that was at to_index.
        }
    if (to_index >= m_len)
        {
        to_index = m_len - 1;
        }
    if (from_index >= m_len)
        {
        from_index = m_len - 1;
        }
    if (from_index == to_index)
        {
        return m_data.get()[to_index];
        }

    // Shift the items between to_index & from_index, and move the item:
    T *copy_src, *copy_dest = m_data.get() + from_index;
    T tmp = *copy_dest;
    do
        {
        copy_src = copy_dest + (from_index < to_index ? 1 : -1);
        *copy_dest = *copy_src;
        copy_dest = copy_src;
        }
        while (copy_src != m_data.get() + to_index);
    *copy_src = tmp;

    // Return a reference to the moved item:
    return *copy_src;
    }


/////////////////////////////////////////////////////////////////////////////
// Path and C-style string helper functions


// Overloaded wrappers for string.h functions:
inline size_t cstrlen( const char *str) { return strlen(str); }
inline size_t cstrlen( const WCHAR *str) { return wcslen(str); }
inline char *cstrcpy( char *dst, const char *src) {return strcpy( dst, src);}
inline WCHAR *cstrcpy( WCHAR *dst, const WCHAR *src) {return wcscpy( dst, src);}
inline char *cstrcat( char *dst, const char *src) {return strcat( dst, src);}
inline WCHAR *cstrcat( WCHAR *dst, const WCHAR *src) {return wcscat( dst, src);}
inline int cstrcmp( const char *s1, const char *s2) {return strcmp( s1, s2);}
inline int cstrcmp( const WCHAR *s1, const WCHAR *s2) {return wcscmp( s1, s2);}
inline int cstricmp(const char *s1, const char *s2) {return stricmp( s1, s2);}
inline int cstricmp(const WCHAR *s1, const WCHAR *s2) {return wcsicmp( s1, s2);}


// These functions copy a 0-terminated string, much like strdup():
template <class CH>
auto_memptr<CH> clone_str( const CH *str)
    {
    auto_memptr<CH> clone =
        auto_memptr<CH>::alloc( (cstrlen( str) + 1) * sizeof(CH));
    cstrcpy( clone.get(), str);
    return clone;
    }


// Returns a new path string: 'parent\child' or 'parentchild'
// (The \ separator is inserted if sep == true.)
template <class CH>
auto_memptr<CH> concat_path( const CH *parent, const CH *child, bool sep = true)
    {
    size_t parent_len = cstrlen( parent);
    size_t child_len = cstrlen( child);
    auto_memptr<CH> path = auto_memptr<CH>::alloc(
        (parent_len + child_len + 2) * sizeof(CH));

    cstrcpy( path.get(), parent);
    if (parent_len && child_len && sep)
        {
        path.get()[parent_len] = '\\';
        path.get()[parent_len + 1] = '\0';
        }
    cstrcat( path.get(), child);
    return path;
    }


// Returns true if the given path exists
inline bool path_exists( const char *path) { return _access( path, 0) == 0; }
inline bool path_exists( const WCHAR *path) { return _waccess( path, 0) == 0; }


// Returns true if the child is a subpath of the parent
bool path_is_child( const TCHAR *parent, const TCHAR *child)
    {
    size_t length = _tcsclen(parent);
    return length < _tcsclen(child) && _tcsnicmp( parent, child, length) == 0;
    }


/////////////////////////////////////////////////////////////////////////////
// Filesystem helpers


// Deletes the file named by the given path.
inline void unlink_file( const TCHAR *path)
    {
    if (DeleteFile( path) == 0)
        {
        throw favlib_win_error();
        }
    }


// Deletes the directory named by the given path, and all its children.
void recursive_unlink( const TCHAR *path)
    {
    WIN32_FIND_DATA found;
    HANDLE handle = FindFirstFile( concat_path( path, _T("*.*")).get(), &found);
    if (handle == INVALID_HANDLE_VALUE)
        {
        throw favlib_win_error();
        }
    do
        {
        if (cstrcmp( found.cFileName, _T(".")) != 0 &&
            cstrcmp( found.cFileName, _T("..")) != 0)
            {
            if (found.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
                {
                recursive_unlink( concat_path( path, found.cFileName).get());
                }
            else
                {
                unlink_file( concat_path( path, found.cFileName).get());
                }
            }
        }
        while (FindNextFile( handle, &found));
    FindClose( handle);

    if (RemoveDirectory( path) == 0)
        {
        throw favlib_win_error();
        }
    }


// Copies the file named by the given path.
// Throws an exception on existing file or other failure.
inline void copy_file( const TCHAR *from_path, const TCHAR *to_path)
    {
    if (CopyFile( from_path, to_path, TRUE) == 0)
        {
        throw favlib_win_error();
        }
    }


// Copies the directory named by the given path, and all its children.
// Throws an exception on existing file or other failure.
void copy_directory( const TCHAR *from_path, const TCHAR *to_path)
    {
    if (path_exists( to_path))
        {
        throw favlib_win_error( ERROR_ALREADY_EXISTS);
        }
    if (path_is_child( from_path, to_path))
        {
        throw favlib_win_error( ERROR_BAD_PATHNAME);
        }

    WIN32_FIND_DATA found;
    HANDLE handle =
        FindFirstFile( concat_path( from_path, _T("*.*")).get(), &found);
    if (handle == INVALID_HANDLE_VALUE)
        {
        throw favlib_win_error();
        }
    if (CreateDirectory( to_path, NULL) == 0)
        {
        throw favlib_win_error();
        }
    do
        {
        if (cstrcmp( found.cFileName, _T(".")) != 0 &&
            cstrcmp( found.cFileName, _T("..")) != 0)
            {
            auto_memptr<TCHAR> from_child =
                concat_path( from_path, found.cFileName);
            auto_memptr<TCHAR> to_child =
                concat_path( to_path, found.cFileName);
            if (found.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
                {
                copy_directory( from_child.get(), to_child.get());
                }
            else
                {
                copy_file( from_child.get(), to_child.get());
                }
            }
        }
        while (FindNextFile( handle, &found));
    FindClose( handle);
    }


// A class for monitoring changes to the filesystem.
class filesystem_change
    {
    public:
        filesystem_change( const WCHAR *path, bool include_subdirs);
        filesystem_change( const filesystem_change &other);
        filesystem_change& operator=( const filesystem_change &other);
        ~filesystem_change() throw() { close(); }

        bool wait( int milliseconds = -1) const; // returns false on timeout

    private:
        filesystem_change(): m_handle(INVALID_HANDLE_VALUE) {}
        void close() throw();
        HANDLE m_handle;
    };


filesystem_change::filesystem_change( const WCHAR *path, bool include_subdirs)
    {
    USES_CONVERSION;

    m_handle = FindFirstChangeNotification(
        W2CT(path),
        include_subdirs ? TRUE : FALSE,
        FILE_NOTIFY_CHANGE_FILE_NAME |
        FILE_NOTIFY_CHANGE_DIR_NAME |
        FILE_NOTIFY_CHANGE_LAST_WRITE);
    if (m_handle == INVALID_HANDLE_VALUE)
        {
        throw favlib_win_error();
        }
    }


filesystem_change::filesystem_change( const filesystem_change &other)
    {
    if (DuplicateHandle( GetCurrentProcess(), other.m_handle,
        GetCurrentProcess(), &m_handle, 0, FALSE, DUPLICATE_SAME_ACCESS) == 0)
        {
        throw favlib_win_error();
        }
    }


filesystem_change& filesystem_change::operator=( const filesystem_change &other)
    {
    close();
    if (DuplicateHandle( GetCurrentProcess(), other.m_handle,
        GetCurrentProcess(), &m_handle, 0, FALSE, DUPLICATE_SAME_ACCESS) == 0)
        {
        throw favlib_win_error();
        }
    return *this;
    }


void filesystem_change::close() throw()
    {
    if (m_handle != INVALID_HANDLE_VALUE)
        {
        FindCloseChangeNotification( m_handle);
        m_handle = INVALID_HANDLE_VALUE;
        }
    }


bool filesystem_change::wait( int milliseconds) const
    {
    DWORD timeout_arg = milliseconds < 0 ? INFINITE: milliseconds;

    // Wait for a change notification
    switch (WaitForSingleObject( m_handle, timeout_arg))
        {
        case WAIT_FAILED:
            throw favlib_win_error();
        case WAIT_TIMEOUT:
            return false;
        }

    // Request future change notifications
    if (FindNextChangeNotification( m_handle) == 0)
        {
        throw favlib_win_error();
        }

    return true;
    }


/////////////////////////////////////////////////////////////////////////////
// LPITEMIDLIST helpers and wrapper class


// These ease walking the parts of an ITEMIDLIST:
union cracked_pidl
    {
    cracked_pidl( LPITEMIDLIST p): pidl(p) {}
    cracked_pidl& operator=( LPITEMIDLIST p)
        { pidl = p; return *this; }
    LPITEMIDLIST pidl;
    UNALIGNED USHORT *sizeptr;
    BYTE *byteptr;
    };
union cracked_const_pidl
    {
    cracked_const_pidl( LPCITEMIDLIST p): pidl(p) {}
    cracked_const_pidl& operator=( LPCITEMIDLIST p)
        { pidl = p; return *this; }
    LPCITEMIDLIST pidl;
    const UNALIGNED USHORT *sizeptr;
    const BYTE *byteptr;
    };


// A class like std::auto_ptr, but for shell-allocated ITEMIDLIST pointers.
// Not for STL containers; copy & assignment steal ownership of internal data.
class auto_pidl: private auto_memptr<ITEMIDLIST>
    {
    public:
        explicit auto_pidl( LPITEMIDLIST pidl = 0) throw()
            : auto_memptr<ITEMIDLIST>(pidl) {}
        auto_pidl( auto_pidl &other) throw()
            : auto_memptr<ITEMIDLIST>(other.disown()) {}
        auto_pidl& operator=( auto_pidl &other) throw()
            { reset( other.disown()); return *this; }

        auto_pidl& append( const auto_pidl &other);

        bool is_zero() const throw() { return get() == 0; }
        bool is_empty() const throw() { return get()->mkid.cb == 0; }
        unsigned size() const throw();

        LPITEMIDLIST get() throw()
            { return auto_memptr<ITEMIDLIST>::get(); }
        LPCITEMIDLIST get() const throw()
            { return auto_memptr<ITEMIDLIST>::get(); }
        LPCITEMIDLIST get_leaf() const throw();
        auto_pidl get_parent() const;

    private:
        explicit auto_pidl( auto_memptr<ITEMIDLIST> &mem) throw()
            : auto_memptr<ITEMIDLIST>(mem) {}
    };


// Returns the size (in chars) of the ITEMIDLIST, including the final 0x0000.
unsigned auto_pidl::size() const
    {
    unsigned total_size = sizeof(USHORT);
    for (cracked_const_pidl p = get(); *p.sizeptr; p.byteptr += *p.sizeptr)
        {
        total_size += *p.sizeptr;
        }
    return total_size;
    }


auto_pidl& auto_pidl::append( const auto_pidl &other)
    {
    unsigned other_size = other.size();
    if (other_size)
        {
        unsigned head_size = size() - sizeof(USHORT);
        resize( head_size + other_size);
        cracked_pidl p = get();
        CopyMemory( p.byteptr + head_size, other.get(), other_size);
        }
    return *this;
    }


// Returns a pointer to the leaf item in our pidl, or 0 if there isn't one.
LPCITEMIDLIST auto_pidl::get_leaf() const throw()
    {
    if (is_zero())
        {
        return 0;
        }

    LPCITEMIDLIST last = 0;
    for (cracked_const_pidl p = get(); *p.sizeptr; p.byteptr += *p.sizeptr)
        {
        last = p.pidl;
        }
    return last;
    }


// Constructs and returns a new pidl representing our pidl's parent.
auto_pidl auto_pidl::get_parent() const
    {
    // determine how many bytes precede our leaf item:
    cracked_const_pidl head = get();
    cracked_const_pidl leaf = get_leaf();
    if (!leaf.pidl)
        {
        return 0;
        }
    unsigned leaf_offset = leaf.byteptr - head.byteptr;

    // Construct a new pidl from the pre-leaf parts, and add the tail 0.
    auto_pidl parent( alloc( leaf_offset + sizeof(USHORT)));
    CopyMemory( parent.get(), get(), leaf_offset);
    cracked_pidl p = parent.get();
    p.byteptr += leaf_offset;
    *p.sizeptr = 0;
    return parent;
    }


/////////////////////////////////////////////////////////////////////////////
// Registry key helper


// Not for STL containers; copy & assignment steal ownership of internal data.
// In general, read failures return false & write failures throw an exception,
// because our calling code expects to encounter missing data regularly.
class auto_hkey
    {
    public:
        explicit auto_hkey( HKEY key = invalid) throw()
            : m_key(key) {}
        auto_hkey( auto_hkey &other) throw()
            : m_key(other.disown()) {}
        auto_hkey& operator=( auto_hkey &other) throw()
            { reset( other.disown()); return *this; }
        auto_hkey( HKEY rootkey, LPCTSTR subkey, bool write = false)
            : m_key(invalid) { open( rootkey, subkey, write); }
        ~auto_hkey() { close(); }

        void reset( HKEY key = invalid)
            { if (m_key != key) { close(); m_key = key; } }

        HKEY get() throw()
            { return m_key; }

        bool open( HKEY rootkey, LPCTSTR subkey, bool write = false);
        void close();

        bool get_value( auto_array<BYTE> &buf, DWORD &type,
            LPCTSTR name, unsigned maxsize = -1);
        bool get_value(
            auto_array<BYTE> &buf, LPCTSTR name, unsigned maxsize = -1)
            { DWORD type; return get_value( buf, type, name, maxsize); }

        void set_value(
            const auto_array<BYTE> &buf, DWORD type, LPCTSTR name);
        void set_value( const auto_array<BYTE> &buf, LPCTSTR name)
            { set_value( buf, REG_BINARY, name); }

        // Thin win32 wrappers: ugly, but efficient in loops.
        inline bool query_info( LPDWORD subkey_count, LPDWORD subkey_name_len,
            LPDWORD value_count, LPDWORD value_name_len, LPDWORD value_size);
        inline bool enum_subkey_name(
            DWORD index, LPTSTR name, LPDWORD name_len);
        inline bool enum_value( DWORD index, LPTSTR name, LPDWORD name_len,
            LPDWORD type, LPBYTE value, LPDWORD value_size);

    private:
        HKEY disown() throw()
            { HKEY tmp = m_key; m_key = invalid; return tmp; }
        bool query_value( DWORD &size, DWORD &type, LPCTSTR name)
            { return m_key != invalid && RegQueryValueEx(
            m_key, name, NULL, &type, NULL, &size) == ERROR_SUCCESS; }

        static const HKEY invalid;
        HKEY m_key;
    };


const HKEY auto_hkey::invalid = (HKEY)INVALID_HANDLE_VALUE;


/*
If write is true, creates subkeys as necessary & opens for write access.
On success, returns true.
On failed open for read, returns false.
On failed open for write, throws favlib_win_error.
*/
bool auto_hkey::open( HKEY rootkey, LPCTSTR subkey, bool write)
    {
    close();
    LONG result;
    // (MSDN shows RegOpenKeyEx() & RegCreateKeyEx() handling "key\\childkey".)
    if (write)
        {
        DWORD junk;
        result = RegCreateKeyEx( rootkey, subkey, 0, NULL,
            REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &m_key, &junk);
        }
    else
        {
        result = RegOpenKeyEx( rootkey, subkey, 0, KEY_READ, &m_key);
        }
    if (result != ERROR_SUCCESS)
        {
        m_key = invalid; // because RegOpenKeyEx() failure sets m_key to 0.
        if (write)
            {
            throw favlib_win_error( result);
            }
        return false;
        }
    return true;
    }


void auto_hkey::close()
    {
    if (m_key != invalid)
        {
        LONG result = RegCloseKey( m_key);
        if (result != ERROR_SUCCESS)
            {
            throw favlib_win_error( result);
            }
        m_key = invalid;
        }
    }


/*
Copies key data into "buf" and type into "type".
Resizes "buf" only if necessary.
On success, returns true.
On failed read, returns false.
If data larger than maxsize, throws favlib_win_error.
Contents of "buf" are not guaranteed to survive a failure.
*/
bool auto_hkey::get_value(
    auto_array<BYTE> &buf, DWORD &type, LPCTSTR name, unsigned maxsize)
    {
    // find out how big the value's data is:
    DWORD size;
    if (m_key == invalid || !query_value( size, type, name))
        {
        return false;
        }
    if (size > maxsize)
        {
        throw favlib_win_error( ERROR_BUFFER_OVERFLOW);
        }

    // retrieve the value's data:
    if (buf.length() < size)
        {
        buf = auto_array<BYTE>( size);
        }
    LONG result = RegQueryValueEx( m_key, name, NULL, NULL, buf.begin(), &size);
    if (result != ERROR_SUCCESS)
        {
        return false;
        }
    return true;
    }


// On failure, throws favlib_win_error.
void auto_hkey::set_value(
    const auto_array<BYTE> &buf, DWORD type, LPCTSTR name)
    {
    if (m_key == invalid)
        {
        return;
        }

    // store the value's data:
    LONG result = RegSetValueEx(
        m_key, name, 0, type, buf.begin(), buf.length());
    if (result != ERROR_SUCCESS)
        {
        throw favlib_win_error( result);
        }
    }


// Thin wrapper for RegQueryInfoKey(); ugly, but efficient in loops.
// Any argument can be NULL if the caller doesn't want that information.
// Returns false on failure.
inline bool auto_hkey::query_info(
    LPDWORD subkey_count, LPDWORD subkey_name_len, LPDWORD value_count,
    LPDWORD value_name_len, LPDWORD value_size)
    {
    return m_key != invalid && RegQueryInfoKey( m_key, NULL, NULL,
        NULL, subkey_count, subkey_name_len, NULL, value_count,
        value_name_len, value_size, NULL, NULL) == ERROR_SUCCESS;
    }


// Thin wrapper for RegEnumKeyEx(); ugly, but efficient in loops.
// Returns false on failure.
inline bool auto_hkey::enum_subkey_name(
    DWORD index, LPTSTR name, LPDWORD name_len)
    {
    FILETIME junk;
    return m_key != invalid && RegEnumKeyEx( m_key, index, name, name_len,
        NULL, NULL, NULL, &junk) == ERROR_SUCCESS;
    }


// Thin wrapper for RegEnumValue(); ugly, but efficient in loops.
// Returns false on failure.
inline bool auto_hkey::enum_value( DWORD index, LPTSTR name, LPDWORD name_len,
    LPDWORD type, LPBYTE value, LPDWORD value_size)
    {
    return m_key != invalid && RegEnumValue( m_key, index, name, name_len,
        NULL, type, value, value_size) == ERROR_SUCCESS;
    }


// Removes the specified registry key, and all its children.
// If the key does not exist, does nothing.
void remove_registry_key( HKEY rootkey, LPCTSTR subkey)
    {
    assert( cstrlen( subkey)); // Try not to remove any major keys, okay?

    auto_hkey doomed_key( rootkey, subkey);
    DWORD child_count;
    DWORD child_name_len;
    if (doomed_key.query_info( &child_count, &child_name_len, NULL, NULL, NULL))
        {
        // Recursively remove any child keys
        auto_array<TCHAR> child_name( child_name_len + 1);
        for (DWORD i = 0; i != child_count; ++i)
            {
            child_name_len = child_name.length();
            if (doomed_key.enum_subkey_name( i,
                child_name.data(), &child_name_len))
                {
                remove_registry_key( doomed_key.get(), child_name.data());
                }
            }

        // Remove the originally specified key:
        doomed_key.close();
        LONG result = RegDeleteKey( rootkey, subkey);
        if (result != ERROR_SUCCESS)
            {
            throw favlib_win_error( result);
            }
        }
    }


/*
Copies key data into "buf", and returns the size of the data.
Resizes "buf" if necessary.
On success, returns true.
On failed read, returns false.
If data larger than maxsize, throws favlib_win_error.
(Contents of "buf" are not guaranteed to survive a failure.)
*/
inline bool get_registry_value( auto_array<BYTE> &buf, LPCTSTR name,
    HKEY rootkey, LPCTSTR subkey, unsigned maxsize = -1)
    {
    auto_hkey key( rootkey, subkey);
    return key.get_value( buf, name, maxsize);
    }


// Stores the contents of "buf" in the named registry value.
// Uses the REG_BINARY type.
// On failure, throws favlib_win_error.
inline void set_registry_value( const auto_array<BYTE> &buf, LPCTSTR name,
    HKEY rootkey, LPCTSTR subkey)
    {
    auto_hkey key( rootkey, subkey, true);
    key.set_value( buf, name);
    }


/*
Copies the specified registry key, and all its children.
On success, returns true.
If the key does not exist, returns false.
On failure, throws favlib_win_error.
*/
bool copy_registry_key( HKEY rootkey, LPCTSTR from_subkey, LPCTSTR to_subkey)
    {
    // Prevent copy of a key into one of its child keys
    if (path_is_child( from_subkey, to_subkey))
        {
        throw favlib_win_error( ERROR_BAD_PATHNAME);
        }

    // Get the number of children, and the lengths of their names & data
    auto_hkey from_key( rootkey, from_subkey);
    DWORD subkey_count,
        subkey_name_len,
        value_count,
        value_name_len,
        value_size;
    if (!from_key.query_info( &subkey_count, &subkey_name_len,
        &value_count, &value_name_len, &value_size))
        {
        return false;
        }
    auto_array<TCHAR> child_name( max(subkey_name_len, value_name_len) + 1);

    // Recursively copy each subkey first
    for (DWORD i = 0; i != subkey_count; ++i)
        {
        // Get a subkey name
        subkey_name_len = child_name.length();
        if (!from_key.enum_subkey_name( i, child_name.data(), &subkey_name_len))
            {
            throw favlib_win_error( ERROR_FILE_NOT_FOUND);
            }

        // Copy the subkey, recursively
        copy_registry_key( rootkey,
            concat_path( from_subkey, child_name.begin()).get(),
            concat_path( to_subkey, child_name.begin()).get());
        }

    // Copy each value
    auto_hkey to_key( rootkey, to_subkey, true);
    auto_array<BYTE> value( value_size);
    DWORD type;
    for (i = 0; i != value_count; ++i)
        {
        // Get a value
        value_name_len = child_name.length();
        value_size = value.length();
        if (!from_key.enum_value( i, child_name.data(), &value_name_len,
            &type, value.data(), &value_size))
            {
            throw favlib_win_error( ERROR_FILE_NOT_FOUND);
            }

        // Copy the value
        to_key.set_value( value, type, child_name.begin());
        }

    return true;
    }


/////////////////////////////////////////////////////////////////////////////
// Shell folder helpers (for IShellFolder and ITEMIDLIST)


typedef auto_ifptr<IShellFolder> folder_iface;
typedef auto_ifptr<IEnumIDList> enum_iface;


// Returns the IShellFolder for the desktop, which is the shell's root folder:
folder_iface get_desktop_iface()
    {
    IShellFolder *desktopif;
    HRESULT err = SHGetDesktopFolder( &desktopif);
    if (err != NOERROR)
        {
        throw favlib_com_error( err);
        }
    return desktopif;
    }


// Returns the IShellFolder for the specified subfolder:
folder_iface get_subfolder_iface(
    folder_iface &folderif, const auto_pidl &subfolder)
    {
    IShellFolder *wantedif;
    HRESULT result = folderif->BindToObject( subfolder.get(), NULL,
        IID_IShellFolder, reinterpret_cast<void**>(&wantedif));
    if (result != S_OK)
        {
        throw favlib_com_error( result);
        }
    return wantedif;
    }


// Returns an auto_pidl for the current user's root favorites folder.
auto_pidl get_favroot_pidl()
    {
    LPITEMIDLIST favs;
    HRESULT err = SHGetSpecialFolderLocation( NULL, CSIDL_FAVORITES, &favs);
    if (err != S_OK)
        {
        throw favlib_com_error( err);
        }
    return favs;
    }


// Returns an auto_pidl relative to the specified folder,
// representing the given file name.
auto_pidl get_relative_pidl( folder_iface &parentif, const WCHAR *fname)
    {
    // convert the child path to an ITEMIDLIST:
    // (Note1: ParseDisplayName() isn't declared with a const input path,
    //  so we copy the file name with clone_str().)
    // (Note2: ParseDisplayName() accepts the for-parsing child item name,
    //  not the for-display name it implies.  For details, see the docs for
    //  IShellFolder::GetDisplayNameOf() and SHGNO Enumerated Type.)
    LPITEMIDLIST child_pidl;
    HRESULT result = parentif->ParseDisplayName(
        NULL, NULL, clone_str(W2COLE(fname)).get(), NULL, &child_pidl, NULL);
    if (FAILED( result))
        {
        throw favlib_com_error( result);
        }

    return child_pidl;
    }


// Returns the IShellFolder for the specified favorites subfolder:
folder_iface get_favfolder_iface( const auto_pidl &subfolder = auto_pidl())
    {
    // get the desired subfolder's full pidl:
    auto_pidl fullpidl = get_favroot_pidl();
    if (!subfolder.is_zero())
        {
        fullpidl.append( subfolder);
        }

    // use the desktop's IShellFolder to create one for the subfolder:
    return get_subfolder_iface( get_desktop_iface(), fullpidl);
    }


// Returns the IShellFolder for the specified favorites subfolder:
folder_iface get_favfolder_iface( const WCHAR *subfolder)
    {
    // get the root favorites folder's IShellFolder:
    folder_iface favif = get_favfolder_iface();
    if (!*subfolder)
        {
        return favif;
        }

    // get the subfolder's IShellFolder:
    return get_subfolder_iface( favif, get_relative_pidl( favif, subfolder));
    }


// Returns an IEnumIDList * for the specified folder.
// Enumerated objects will not include files unless "with_files" is true.
enum_iface get_enum_iface( folder_iface &folderif, bool with_files)
    {
    DWORD flags = SHCONTF_FOLDERS | (with_files ? SHCONTF_NONFOLDERS : 0);
    IEnumIDList *enumif;
    HRESULT result = folderif->EnumObjects( NULL, flags, &enumif);
    if (result != S_OK)
        {
        throw favlib_com_error( result);
        }
    return enumif;
    }


/////////////////////////////////////////////////////////////////////////////
// A wrapper class for shell object attributes.


class shell_attribs
    {
    public:
        shell_attribs() throw(): m_attribs(0) {}
        // (this class uses default member-by-member copy & assignment)
        shell_attribs( folder_iface &folderif, LPCITEMIDLIST pidl)
            { init( folderif, pidl); }
        ~shell_attribs() throw() {}

        void init() { m_attribs = 0; }
        void init( folder_iface &folderif, LPCITEMIDLIST pidl)
            { m_attribs = get( folderif, pidl); }

        bool is_folder() const throw()
            { return (m_attribs & SFGAO_FOLDER) != 0; }
        bool has_subfolder() const throw()
            { return (m_attribs & SFGAO_HASSUBFOLDER) != 0; }

        static bool is_folder( folder_iface &folderif, LPCITEMIDLIST pidl)
            { return (get( folderif, pidl, SFGAO_FOLDER) & SFGAO_FOLDER) != 0; }

    private:
        static ULONG get( folder_iface &folderif, LPCITEMIDLIST pidl,
            ULONG flags = SFGAO_FOLDER | SFGAO_HASSUBFOLDER);

        ULONG m_attribs;
    };


// Returns the specified windows attribute flags for the given pidl.
ULONG shell_attribs::get(
    folder_iface &folderif, LPCITEMIDLIST pidl, ULONG flags)
    {
    ULONG attrs = flags;
    HRESULT result = folderif->GetAttributesOf( 1, &pidl, &attrs);
    if (result != S_OK)
        {
        throw favlib_com_error( result);
        }
    return attrs;
    }


// Returns attributes of the root favorites folder.
shell_attribs get_favroot_attribs()
    {
    auto_pidl root = get_favroot_pidl();
    folder_iface parentif = get_subfolder_iface(
        get_desktop_iface(), root.get_parent());
    return shell_attribs(parentif,root.get_leaf());
    }


/////////////////////////////////////////////////////////////////////////////
// A wrapper class (and helper functions) for shell object names.


// Helper function for shell_name class.
// Retrieves a shell object's parse or display name, in Windows STRRET format.
// Returns the absolute path if full_path is true and for_display is false.
void get_pidl_name( STRRET &name, folder_iface &folderif, LPCITEMIDLIST pidl,
    bool for_display, bool full_path)
    {
    assert( !(for_display && full_path));

    // Retrieve the item's name string, in whichever format Windows chooses:
    name.uType = for_display ? STRRET_OFFSET : STRRET_WSTR; // hint to Windows
    DWORD format = (full_path ? 0 : SHGDN_INFOLDER) |
        (for_display ? 0 : SHGDN_FORPARSING);
    HRESULT err = folderif->GetDisplayNameOf( pidl, format, &name);
    if (err != S_OK)
        {
        throw favlib_com_error( err);
        }
    }


// Helper function for shell_name class.
// Returns a shell object's name, suitable for parsing.
// Returns the absolute path if full_path is true.
auto_memptr<WCHAR> get_pidl_parse_name(
    folder_iface &folderif, LPCITEMIDLIST pidl, bool full_path)
    {
    STRRET s;
    get_pidl_name( s, folderif, pidl, false, full_path);

    USES_CONVERSION;
    const WCHAR *str;
    switch (s.uType)
        {
        case STRRET_WSTR:
            return s.pOleStr;
        case STRRET_OFFSET:
            str = A2CW( reinterpret_cast<const char *>(pidl) + s.uOffset);
            break;
        case STRRET_CSTR:
            str = A2CW( s.cStr);
            break;
        default:
            str = L"";
            break;
        }
    auto_memptr<WCHAR> name =
        auto_memptr<WCHAR>::alloc( (cstrlen( str) + 1) * sizeof(WCHAR));
    cstrcpy( name.get(), str);
    return name;
    }


// Helper function for shell_name class.
// Returns a shell object's name, suitable for display.
auto_memptr<TCHAR> get_pidl_display_name(
    folder_iface &folderif, LPCITEMIDLIST pidl)
    {
    STRRET s;
    get_pidl_name( s, folderif, pidl, true, false);

    const TCHAR *str;
    USES_CONVERSION;
    switch (s.uType)
        {
        case STRRET_WSTR:
            str = W2CT( s.pOleStr);
            break;
        case STRRET_OFFSET:
            str = A2CT( reinterpret_cast<const char *>(pidl) + s.uOffset);
            break;
        case STRRET_CSTR:
            str = A2CT( s.cStr);
            break;
        default:
            str = _T("");
            break;
        }
    auto_memptr<TCHAR> name =
        auto_memptr<TCHAR>::alloc( (cstrlen( str) + 1) * sizeof(TCHAR));
    cstrcpy( name.get(), str);
    if (s.uType == STRRET_WSTR)
        {
        com_free( s.pOleStr);
        }
    return name;
    }


// This class wraps a shell object's name, in multiple formats.
// (Using a class for this will let us optimize memory allocation later.)
// Not for STL containers; copy & assignment steal ownership of internal data.
class shell_name
    {
    public:
        shell_name() {}
        // (this class uses default member-by-member copy & assignment)
        shell_name( folder_iface &folderif, LPCITEMIDLIST pidl)
            { assign( folderif, pidl); }
        ~shell_name() {}

        void assign( folder_iface &folderif, LPCITEMIDLIST pidl)
            { m_parse = get_pidl_parse_name( folderif, pidl, false);
            m_display = get_pidl_display_name( folderif, pidl); }
        void erase() { m_parse.reset(); m_display.reset(); }

        const WCHAR *for_parse() const throw()
            { return m_parse.is_zero() ? L"" : m_parse.get(); }
        const TCHAR *for_display() const throw()
            { return m_display.is_zero() ? _T("") : m_display.get(); }

    private:
        auto_memptr<WCHAR> m_parse;
        auto_memptr<TCHAR> m_display;
    };


// Returns the absolute path of a favorites file/folder,
// given a path relative to the root favorites folder.
auto_memptr<WCHAR> get_fav_full_path( const WCHAR *rel_path = L"")
    {
    auto_memptr<WCHAR> path = get_pidl_parse_name(
        get_desktop_iface(), get_favroot_pidl().get(), true);
    if (*rel_path != 0)
        {
        path = concat_path( path.get(), rel_path);
        }
    return path;
    }


/////////////////////////////////////////////////////////////////////////////
// MenuOrder registry data helper


/*
Windows allows users to reorder the contents of a favorites folder,
and stores that order in the registry.  The registry keys and values
involved are fairly well known, but (as of 2004) the data format is
not officially documented.  The following describes the data layout
and format.

Thanks go to the authors of these three articles, each of which has
incomplete and/or incorrect parts, but all of which were helpful:

    http://sourceforge.net/mailarchive/message.php?msg_id=1088896
    http://www.arstdesign.com/articles/iefavorites.html
    http://www.experts-exchange.com/Operating_Systems/Win2000/Q_20618339.html

Favorites are stored as .url files in the file system.  Each user has
his own favorites folder, recognized by the shell (Explorer) as special,
and locatable using an API call.  A typical favorites folder path is:

    C:\Documents and Settings\Username\Favorites

When the user opens the Favorites menu in IE or Windows Explorer, a
menu structure mirroring the contents of that folder appears.  Each
file appears as a menu item, and each subfolder appears as a submenu.
The user can arrange the contents of any favorites menu in any order
he pleases, using drag & drop.  Windows provides the IShellMenu
interface for adding such a menu to an application's GUI, but it
doesn't seem to expose much programmatic control of the underlying
data.

User-customized ordering information is stored in the registry, using
a key structure that mirrors the favorites folder structure, starting
with this key:

    HKEY_CURRENT_USER\
    Software\Microsoft\Windows\CurrentVersion\Explorer\MenuOrder\Favorites

Each key contains a single REG_BINARY value named "Order", whose data
represents the custom ordering of items within the corresponding folder.

Whenever a favorites menu is displayed, Windows presents its contents
in the following order:

    1. Files & folders with custom ordering as described in the registry.
    2. Remaining folders, in alphabetical order.
    3. Remaining files, in alphabetical order.

The display order data format somewhat resembles the Electronic Arts
IFF 85 interchange file format.  It is a sequence of two data blocks
(each prefixed with size and id), the second of which contains a
sequence of Windows ITEMIDLIST structures (each prefixed with size and
display position and postfixed by a semi-mysterious bitfield).  The
ITEMIDLIST structures, which account for the bulk of the data, are
basically just complicated file system paths for use by the Windows
shell.  In other words, an application does not have to understand
their internals, but can instead use Windows shell APIs to manipulate
them.

For the curious: the contents of a common ITEMIDLIST include Unicode
paths, ANSI character paths, last-modified times (in DOS/FAT format),
etc.  You might find other things within an ITEMID that was generated
by a shell namespace extension.

A few things to keep in mind when working with Windows Shell paths:
- The Windows API uses a type called a PIDL, which means "pointer
  to ITEMIDLIST".
- PIDLS generally point to memory that has been allocated by the
  Windows shell, so it may be unwise to use a different allocator
  when passing PIDLs to Windows functions.
- The IShellFolder OLE/COM/ActiveX interface can be very helpful.
- Many of the most convenient Windows API functions and interfaces
  are not present in older versions of 32-bit Windows, such as
  Windows 95 or NT.  One must be careful to avoid using them when
  backward compatibility is important.

*/


/*
A helper class for retrieving the order of items in a favorites folder.
It holds a copy of the raw ordering data from the registry, and offers
accessors for reading that data in a way similar to using an STL container.
We use a class to tie a pidl pointer array to the raw data, instead of
passing a pidl array between functions, so the pidl contents remain within
the raw ordering data instead of being individually allocated copies.
*/
class favs_order
    {
    public:
        favs_order() throw() {}
        // (this class uses default member-by-member copy & assignment)
        explicit favs_order( const TCHAR *subfolder) { load( subfolder); }
        ~favs_order() throw() {}

        typedef const LPCITEMIDLIST *const_iterator;

        const_iterator begin() const throw() { return m_pidls.begin(); }
        const_iterator end() const throw() { return m_pidls.end(); }

        void erase() { m_pidls.erase(); m_data.erase(); }
        void load( const TCHAR *subfolder)
            { erase(); if (fetch_data( subfolder)) { parse_data(); } }

        static void save( const TCHAR *subfolder,
            folder_iface &folderif, const auto_array<auto_pidl> &pidls);
        static bool copy(
            const TCHAR *from_subfolder, const TCHAR *to_subfolder)
            { return copy_registry_key( HKEY_CURRENT_USER,
            concat_path( root_key, from_subfolder).get(),
            concat_path( root_key, to_subfolder).get()); }
        static void remove( const TCHAR *subfolder)
            { remove_registry_key( HKEY_CURRENT_USER,
            concat_path( root_key, subfolder).get()); }

    private:
        bool fetch_data( const TCHAR *subfolder);
        void parse_data();

        static auto_array<BYTE> build_data(
            folder_iface &folderif, const auto_array<auto_pidl> &pidls);

        static const unsigned maxsize;
        static const DWORD head_block_id;
        static const DWORD order_block_id;
        static const TCHAR *root_key;

        // These represent section headers within the registry data:
        struct generic_block_header
            {
            DWORD size;
            DWORD id;
            };
        struct order_block_header
            {
            DWORD size;
            DWORD id;
            DWORD numrecs;
            };
        struct order_record
            {
            DWORD size;
            DWORD sortpos;
            ITEMIDLIST idlist;
            };
        struct order_block
            {
            order_block_header hdr;
            order_record rec;
            };

        // These ease walking the sections within the raw registry data:
        union cracked_order_record
            {
            const UNALIGNED order_record *recptr;
            const BYTE *byteptr;
            };
        union cracked_block
            {
            const UNALIGNED generic_block_header *genericptr;
            const UNALIGNED order_block *orderptr;
            const BYTE *byteptr;
            };

        // These throw bad_order_data if data is structured incorrectly:
        class bad_order_data: public favlib_exception {};
        void validate_generic_block( const cracked_block &block) const;
        void validate_order_block( const cracked_block &block) const;
        void validate_record( const cracked_order_record &record) const;

        auto_array<BYTE> m_data;
        auto_array<LPCITEMIDLIST> m_pidls;
    };


// Size (in bytes) of the largest registry value we will accept as valid:
const unsigned favs_order::maxsize = 1024 * 1024 * 8;

// Block ID numbers found in all registry data created by Windows:
const DWORD favs_order::head_block_id = 2;
const DWORD favs_order::order_block_id = 1;

// The root registry key for favorites ordering information:
const TCHAR *favs_order::root_key = _T("Software\\Microsoft\\Windows\\")
    _T("CurrentVersion\\Explorer\\MenuOrder\\Favorites");


// Copies the given DWORD value to *ptr, and advances ptr by sizeof(DWORD).
inline void copy_dword( BYTE* &ptr, DWORD value)
    {
    *reinterpret_cast<UNALIGNED DWORD*>(ptr) = value;
    ptr += sizeof(DWORD);
    }


/*
Builds a block of ordering data suitable for saving in the registry,
given a list of pidls.
*/
auto_array<BYTE> favs_order::build_data(
    folder_iface &folderif, const auto_array<auto_pidl> &pidls)
    {
    // Determine the size of the data we're building:
    unsigned data_size =
        sizeof(generic_block_header) + sizeof(order_block_header);
    const unsigned record_overhead = 3 * sizeof(DWORD);
    const auto_pidl *pidl;
    for (pidl = pidls.begin(); pidl != pidls.end(); ++pidl)
        {
        unsigned pidl_size = pidl->size();
        data_size += pidl_size + pidl_size % 2 + record_overhead;
        }

    // Allocate a block of data:
    auto_array<BYTE> data_block( data_size);
    BYTE *data = data_block.begin();

    // Add the first head ID block:
    copy_dword( data, sizeof(generic_block_header));
    copy_dword( data, head_block_id);

    // Add the order block header:
    copy_dword( data, data_size - sizeof(generic_block_header));
    copy_dword( data, order_block_id);
    copy_dword( data, pidls.length());

    // Add an order record for each pidl:
    unsigned position = 0;
    for (pidl = pidls.begin(); pidl != pidls.end(); ++pidl)
        {
        // Add the record's size & sort position:
        unsigned pidl_size = pidl->size();
        copy_dword( data, pidl_size + pidl_size % 2 + record_overhead);
        copy_dword( data, position++);
        // Add the record's ITEMIDLIST & (if necessary) a padding byte:
        memcpy( data, pidl->get(), pidl_size);
        data += pidl_size;
        if (pidl_size % 2)
            {
            *data++ = 0;
            }
        // Add the record's mysterious tail flags:
        DWORD flags = shell_attribs::is_folder( folderif, pidl->get()) ?
            0x0000000D : 0x00000005; //xxx todo: verify these values in winxp
        copy_dword( data, flags);
        }
    assert( data == data_block.end());
    return data_block;
    }


/*
Builds the given favorites subfolder's ordering data from the given
pidls, and writes it to the registry.
*/
void favs_order::save( const TCHAR *subfolder,
    folder_iface &folderif, const auto_array<auto_pidl> &pidls)
    {
    auto_array<BYTE> data_block = build_data( folderif, pidls);
    auto_memptr<TCHAR> subkey = concat_path( root_key, subfolder);
    set_registry_value(
        data_block, _T("Order"), HKEY_CURRENT_USER, subkey.get());
    }


/*
Reads the given favorites subfolder's raw ordering data from the registry.
Overwrites contents of m_data, reallocating if m_data is too small.
Returns false if ordering data is not found in the registry.
*/
bool favs_order::fetch_data( const TCHAR *subfolder)
    {
    auto_memptr<TCHAR> subkey = concat_path( root_key, subfolder);
    return get_registry_value(
        m_data, _T("Order"), HKEY_CURRENT_USER, subkey.get(), maxsize);
    }


/*
Parses the raw ordering data in m_data.  (Call fetch_data() first.)
Fills m_pidls with pointers to items found in the data, in the order
described by the data.  Any unused positions in the sequence will be
null pointers.  (These are caused by items found in the data with
out-of-range sort positions.)
Leaves m_pidls empty if errors are found in the data.
*/
void favs_order::parse_data()
    {
    try
        {
        cracked_block block;
        block.byteptr = m_data.begin();

        // First section of data: generic block with id == 0x00000002
        validate_generic_block( block);
        if (block.genericptr->id != head_block_id)
            {
            throw bad_order_data();
            }

        block.byteptr += block.genericptr->size;

        // Second section of data: ordering info block with id == 0x00000001
        validate_order_block( block);
        const unsigned numrecs = block.orderptr->hdr.numrecs;

        if (block.orderptr->hdr.id != order_block_id)
            {
            throw bad_order_data();
            }

        cracked_order_record rec;
        rec.recptr = &block.orderptr->rec;

        // Save a pidl for each record whose sort position is in range:
        m_pidls = auto_array<LPCITEMIDLIST>( numrecs, 0);
        for (unsigned n = 0; n != numrecs; ++n)
            {
            validate_record( rec);
            if (rec.recptr->sortpos >= 0 && rec.recptr->sortpos < numrecs)
                {
                m_pidls[rec.recptr->sortpos] = &rec.recptr->idlist;
                }
            rec.byteptr += rec.recptr->size;
            }
        }
    catch (bad_order_data)
        {
        m_pidls.erase();
        }
    catch (...)
        {
        m_pidls.erase();
        throw;
        }
    }


// Throws if it's unsafe to treat block as a generic_block_header struct:
void favs_order::validate_generic_block( const cracked_block &block) const
    {
    const BYTE *header_end = block.byteptr + sizeof(generic_block_header);

    if (header_end > m_data.end())
        {
        throw bad_order_data();
        }

    const BYTE *claimed_end = block.byteptr + block.genericptr->size;

    if (claimed_end > m_data.end() || claimed_end < header_end)
        {
        throw bad_order_data();
        }
    }

// Throws if it's unsafe to treat block as an order_block struct:
void favs_order::validate_order_block( const cracked_block &block) const
    {
    const BYTE *header_end = block.byteptr + sizeof(order_block_header);

    if (header_end > m_data.end())
        {
        throw bad_order_data();
        }

    const UNALIGNED order_block_header *header = &block.orderptr->hdr;
    const BYTE *claimed_end = block.byteptr + header->size;

    if (claimed_end > m_data.end() || claimed_end < header_end ||
        (header->numrecs && header->size < sizeof(order_block)))
        {
        throw bad_order_data();
        }
    }

// Throws if it's unsafe to treat record as an order_record struct:
void favs_order::validate_record( const cracked_order_record &record) const
    {
    // check the record size:
    const BYTE *struct_end = record.byteptr + sizeof(order_record);

    if (struct_end > m_data.end())
        {
        throw bad_order_data();
        }

    const BYTE *rec_claimed_end = record.byteptr + record.recptr->size;

    if (rec_claimed_end > m_data.end() || rec_claimed_end < struct_end)
        {
        throw bad_order_data();
        }

    // check the first ITEMID:
    cracked_const_pidl idlist = &record.recptr->idlist;
    const BYTE *itemid_claimed_end = idlist.byteptr + *idlist.sizeptr;

    if (itemid_claimed_end > rec_claimed_end ||
        itemid_claimed_end < struct_end)
        {
        throw bad_order_data();
        }

    // check that the second ITEMID is a 0: (single-level ITEMIDLIST)
    idlist.byteptr += *idlist.sizeptr;
    const BYTE *itemid_struct_end = idlist.byteptr + sizeof(USHORT);

    if (itemid_struct_end > rec_claimed_end || *idlist.sizeptr)
        {
        throw bad_order_data();
        }
    }


/////////////////////////////////////////////////////////////////////////////
// Wrapper class for Windows Internet Shortcuts (.url files)


/*
MS internet shortcuts are stored as files with a .url file name suffix.
They appear to be formatted like .ini files, but their content is not
fully documented.  We use the IUniformResourceLocator interface, as
described here:

http://msdn.microsoft.com/library/default.asp?url=/library/en-us/shellcc/platform/shell/programmersguide/shell_int/shell_int_programming/shortcuts/internet_shortcuts.asp

IUniformResourceLocator seems to work well in practice, but I did notice
some strange behavior when using it to modify an internet shortcut created
by Internet Explorer.  The original IE .url file looked like this:

    [DEFAULT]
    BASEURL=http://example.com/
    [InternetShortcut]
    URL=http://example.com/
    Modified=C0E308E01E46C601B6

After using IUniformResourceLocator to modify the URL, it looked like this:

    [DEFAULT]
    BASEURL=http://example.com/
    [InternetShortcut]
    URL=http://newurl.microsoft.com/
    Modified=60E16CEA2B46C601CF

Notice how the [InternetShortcut] section was updated, but the [DEFAULT]
section was not.  It still contained the old URL.

I spent a fair amount of time trying to access that old URL using the
IPropertySetStorage interface with FMTID_Intshcut and FMTID_InternetSite
property formats, with no luck.

This doesn't seem to be a problem in practice, because IE and Explorer
ignore the old URL.  I suppose some older version of the Windows shell
used [DEFAULT] BASEURL, and IE still creates it for backward compatibility
when .url files are copied between machines.
*/


// An object to contain the data in a Windows Internet Shortcut (.url file).
// Not for STL containers; copy & assignment steal ownership of internal data.
// TODO: expose the shortcut's icons, descriptions, framesets, etc.
class inet_shortcut_cache
    {
    public:
        inet_shortcut_cache() {}
        // (this class uses default member-by-member copy & assignment)
        ~inet_shortcut_cache() {}

        bool is_empty() const { return m_url.is_zero(); }
        void load( const WCHAR *file_path);
        void save( const WCHAR *file_path) const;

        const TCHAR *get_url() const
            { return m_url.is_zero() ? _T("") : m_url.get(); }
        void set_url( const TCHAR *url)
            { m_url = clone_str( url);  }
        // (TODO: cache properties, accessible through IPropertySetStorage)

    private:
        auto_memptr<TCHAR> m_url;
    };


// Returns an empty IUniformResourceLocator object.
auto_ifptr<IUniformResourceLocator> new_url_iface()
    {
    IUniformResourceLocator *ptr;
    HRESULT result = CoCreateInstance(
        CLSID_InternetShortcut, NULL, CLSCTX_INPROC_SERVER,
        IID_IUniformResourceLocator, reinterpret_cast<void**>(&ptr));
    if (FAILED( result))
        {
        throw favlib_com_error( result);
        }

    return ptr;
    }


// Saves data to the specified internet shortcut (.url) file.
void inet_shortcut_cache::save( const WCHAR *file_path) const
    {
    USES_CONVERSION;
    auto_ifptr<IUniformResourceLocator> url_iface = new_url_iface();

    // Set the URL:
    HRESULT result = url_iface->SetURL( is_empty() ? _T("") : m_url.get(),
        IURL_SETURL_FL_USE_DEFAULT_PROTOCOL);
    if (FAILED( result))
        {
        throw favlib_com_error( result);
        }

    // Save the shortcut to disk:
    auto_ifptr<IPersistFile> file =
        query_iface<IPersistFile>( url_iface, IID_IPersistFile);
    result = file->Save( W2COLE(file_path), TRUE);
    if (FAILED( result))
        {
        throw favlib_com_error( result);
        }
    }


// Loads data from the specified internet shortcut (.url) file.
void inet_shortcut_cache::load( const WCHAR *file_path)
    {
    USES_CONVERSION;
    auto_ifptr<IUniformResourceLocator> url_iface = new_url_iface();

    // Read shortcut data from disk:
    auto_ifptr<IPersistFile> file =
        query_iface<IPersistFile>( url_iface, IID_IPersistFile);
    HRESULT result = file->Load( W2COLE(file_path), 0);
    if (FAILED( result))
        {
        throw favlib_com_error( result);
        }

    // Extract the URL:
    LPTSTR url;
    result = url_iface->GetURL( &url);
    if (FAILED( result))
        {
        throw favlib_com_error( result);
        }
    m_url.reset( (result == S_FALSE || !url) ? 0 : url);
    }


/////////////////////////////////////////////////////////////////////////////
// Class representing an item within a favorites folder


// An item within a favorites folder.  (Normally a .url file or a subfolder.)
// Not for STL containers; copy & assignment steal ownership of internal data.
class fav_item: public fav_folder::sortable_item
    {
    public:
        fav_item() {}
        // (this class uses default member-by-member copy & assignment)
        fav_item( folder_iface &folderif, LPCITEMIDLIST pidl)
            : m_name( folderif, pidl), m_attribs( folderif, pidl) {}

        virtual const TCHAR *get_name() const // used by fav_folder::sort()
            { return m_name.for_display(); }
        const WCHAR *get_parse_name() const
            { return m_name.for_parse(); }

        virtual bool is_folder() const
            { return m_attribs.is_folder(); }
        virtual bool has_subfolder() const // used by fav_folder::sort()
            { return m_attribs.has_subfolder(); }
        const shell_attribs &get_attribs() const
            { return m_attribs; }

        const TCHAR *get_url( const WCHAR *fav_subfolder)
            { fill_cache( fav_subfolder); return m_cache.get_url(); }
        void set_url( const WCHAR *fav_subfolder, const TCHAR *url)
            { fill_cache( fav_subfolder); m_cache.set_url( url);
            m_cache.save( get_full_path( fav_subfolder).get()); }

    private:
        auto_memptr<WCHAR> get_full_path( const WCHAR *fav_subfolder) const
            { return get_fav_full_path(
                concat_path( fav_subfolder, get_parse_name()).get()); }
        void fill_cache( const WCHAR *fav_subfolder)
            { if (m_cache.is_empty() && !is_folder())
                { m_cache.load( get_full_path( fav_subfolder).get()); } }

        shell_name m_name;
        shell_attribs m_attribs;
        inet_shortcut_cache m_cache;
    };



/////////////////////////////////////////////////////////////////////////////
// Helper classes & functions for retrieving sorted contents of a folder.


// An STL-style binary predicate class for comparing two single-level pidls.
class pidl_lessthan
    {
    public:
        explicit pidl_lessthan( folder_iface &folderif)
            : m_folderif(folderif.copy()) {}
        pidl_lessthan( const pidl_lessthan &other)
            : m_folderif(other.m_folderif.copy()) {}
        pidl_lessthan& operator=( const pidl_lessthan &other)
            { m_folderif = other.m_folderif.copy(); }
        ~pidl_lessthan() {}

        bool operator()( LPCITEMIDLIST x, LPCITEMIDLIST y)
            { HRESULT result = m_folderif->CompareIDs( 0, x, y);
            if (FAILED( result)) throw favlib_com_error( result);
            return static_cast<short>(HRESULT_CODE(result)) < 0; }

    private:
        pidl_lessthan(); // there is no default constructor

        folder_iface m_folderif;
    };


// A class for retrieving & sorting the contents of a favorites folder.
class folder_contents
    {
    public:
        folder_contents() throw() {}
        folder_contents( folder_iface &folderif, bool with_files)
            { gather( folderif, with_files); }
        ~folder_contents() throw() { free_pidls(); }

        typedef const LPCITEMIDLIST *const_iterator;

        void gather( folder_iface &folderif, bool with_files);
        void erase() throw() { free_pidls(); m_pidls.erase(); }

        unsigned length() const throw() { return m_pidls.length(); }
        const_iterator begin() const throw() { return m_pidls.begin(); }
        const_iterator end() const throw() { return m_pidls.end(); }
        const_iterator find( LPCITEMIDLIST target);
        shell_attribs get_attribs( const_iterator which)
            { return shell_attribs( m_folderif, *which); }

    private:
        // copy and assignment are inefficient, unnecessary, and unimplemented:
        folder_contents( const folder_contents &);
        folder_contents& operator=( const folder_contents &);

        void free_pidls() throw()
            { for_each( m_pidls.begin(), m_pidls.end(), com_free<ITEMIDLIST>); }
        void append_pidls( enum_iface &enumif);

        folder_iface m_folderif;
        auto_array<LPITEMIDLIST>m_pidls;
    };


folder_contents::const_iterator folder_contents::find( LPCITEMIDLIST findme)
    {
    pidl_lessthan isless(m_folderif);
    const LPCITEMIDLIST *found = lower_bound( begin(), end(), findme, isless);
    if (found == end() || isless( findme, *found))
        {
        return end();
        }
    return found;
    }


void folder_contents::gather( folder_iface &folderif, bool with_files)
    {
    m_folderif = folderif.copy();
    erase();
    append_pidls( get_enum_iface( m_folderif, with_files));
    sort( m_pidls.begin(), m_pidls.end(), pidl_lessthan(m_folderif));
    }


void folder_contents::append_pidls( enum_iface &enumif)
    {
    const unsigned buflen = 100; // fetch this many at a time
    LPITEMIDLIST buf[buflen];
    HRESULT result;
    do
        {
        ULONG numfetched;
        result = enumif->Next( buflen, buf, &numfetched);
        if (SUCCEEDED( result))
            {
            unsigned oldlen = m_pidls.length();
            m_pidls.resize( oldlen + numfetched);
            copy( buf, buf + numfetched, m_pidls.begin() + oldlen);
            }
        }
        while (result == NOERROR);
    if (FAILED( result))
        {
        throw favlib_com_error( result);
        }
    }


// Returns the contents of the specified Favorites folder,
// arranged in the same order they would be shown by Internet Explorer.
auto_array<fav_item> get_fav_children( const WCHAR *subfolder, bool with_files)
    {
    USES_CONVERSION;

    // Retrieve the subfolder's contents & their display order:
    favs_order order( W2CT(subfolder));
    folder_iface folderif = get_favfolder_iface( subfolder);
    folder_contents contents( folderif, with_files);
    vector<bool> contents_processed( contents.length(), false);

    // Prepare an array to hold the sorted names of each subitem:
    auto_array<fav_item> out_favs(contents.length());
    unsigned out_offset = 0;

    // Process the ordered items:
    favs_order::const_iterator ordered;
    folder_contents::const_iterator child;
    for (ordered = order.begin(); ordered != order.end(); ++ordered)
        {
        if (*ordered)
            {
            child = contents.find( *ordered);
            if (child != contents.end())
                {
                out_favs[out_offset] = fav_item( folderif, *child);
                ++out_offset;
                contents_processed[child - contents.begin()] = true;
                }
            }
        }

    // Process the remaining folders:
    for (child = contents.begin(); child != contents.end(); ++child)
        {
        if (!contents_processed[child - contents.begin()])
            {
            if (shell_attribs::is_folder( folderif, *child))
                {
                out_favs[out_offset] = fav_item( folderif, *child);
                ++out_offset;
                contents_processed[child - contents.begin()] = true;
                }
            }
        }

    // Process the remaining items (all of which are non-folders):
    for (child = contents.begin(); child != contents.end(); ++child)
        {
        if (!contents_processed[child - contents.begin()])
            {
            out_favs[out_offset] = fav_item( folderif, *child);
            ++out_offset;
            }
        }

    return out_favs;
    }


// Saves the order of the given favorites to the registry.
void save_favfolder_order(
    const WCHAR *subfolder, const auto_array<fav_item> &favs)
    {
    USES_CONVERSION;

    // Build a temporary array of our children's pidls:
    folder_iface folderif = get_favfolder_iface( subfolder);
    auto_array<auto_pidl> pidls( favs.length());
    for (unsigned n = 0; n != favs.length(); ++n)
        {
        pidls[n] = get_relative_pidl( folderif, favs[n].get_parse_name());
        }

    // Save the order of the pidls to the registry:
    favs_order::save( W2CT(subfolder), folderif, pidls);
    }


} // end of private namespace


/////////////////////////////////////////////////////////////////////////////
// fav_folder - the implementation behind the api


//xxx todo: implement copy & assignment behavior for this class
struct fav_folder::private_data
    {
    // Default constructor; initialzes to the favorites root folder.
    explicit private_data( bool with_files = true)
        : attribs(get_favroot_attribs()), subfolder(clone_str( L""))
        { favs = get_fav_children( subfolder.get(), with_files); }

    // Alternate constructor: initialize to a subfolder.
    private_data(
        const private_data &parent, unsigned child_index, bool with_files);

    ~private_data() {}

    folder_iface get_iface()
        { return get_favfolder_iface( subfolder.get()); }
    const WCHAR *get_parse_name( unsigned index) const
        { return favs[index].get_parse_name(); }
    auto_memptr<WCHAR> get_relative_path( const WCHAR *child_name) const
        { return concat_path( subfolder.get(), child_name); }
    auto_memptr<WCHAR> get_relative_path( unsigned index) const
        { return get_relative_path( get_parse_name( index)); }
    auto_memptr<WCHAR> get_full_path() const
        { return get_fav_full_path( subfolder.get()); }
    auto_memptr<WCHAR> get_full_path( const WCHAR *child_name) const
        { return get_fav_full_path(
            concat_path( subfolder.get(), child_name).get()); }
    auto_memptr<WCHAR> get_full_path( unsigned index) const
        { return get_full_path( get_parse_name( index)); }
    auto_pidl get_pidl( unsigned index)
        { return get_relative_pidl( get_iface(), get_parse_name( index)); }
    void save_order() const
        { save_favfolder_order( subfolder.get(), favs); }

    shell_attribs attribs;
    auto_memptr<WCHAR> subfolder; // path relative to favorites root
    auto_array<fav_item> favs; // favorites & subfolders in this folder
    };


fav_folder::private_data::private_data(
    const private_data &parent, unsigned child_index, bool with_files)
    {
    const fav_item &child = parent.favs[child_index];
    attribs = child.get_attribs();
    subfolder = concat_path(
        parent.subfolder.get(), child.get_parse_name());
    favs = get_fav_children( subfolder.get(), with_files);
    }


// Default constructor; initialzes to the favorites root folder.
fav_folder::fav_folder( bool with_files)
    :m(new private_data(with_files))
    {
    }


// Alternate constructor; initializes to a subfolder.
fav_folder::fav_folder(
    const fav_folder &parent, unsigned child_index, bool with_files)
    :m(new private_data(*parent.m, child_index, with_files))
    {
    }


fav_folder::~fav_folder() {}


bool fav_folder::has_subfolder() const
    {
    return m->attribs.has_subfolder();
    }


unsigned fav_folder::length() const
    {
    return m->favs.length();
    }


bool fav_folder::is_folder( unsigned index) const
    {
    return m->favs[index].is_folder();
    }


bool fav_folder::has_subfolder( unsigned index) const
    {
    return m->favs[index].has_subfolder();
    }


const TCHAR *fav_folder::get_name( unsigned index) const
    {
    return m->favs[index].get_name();
    }


const TCHAR *fav_folder::get_url( unsigned index)
    {
    return m->favs[index].get_url( m->subfolder.get());
    }


void fav_folder::set_url( unsigned index, const TCHAR *url)
    {
    m->favs[index].set_url( m->subfolder.get(), url);
    }


// Changes a child item's display name.
void fav_folder::rename( unsigned index, const TCHAR *new_name)
    {
    USES_CONVERSION;

    // Rename the folder or file in the filesystem
    LPITEMIDLIST tmp_pidl = 0;
    folder_iface folderif = m->get_iface();
    HRESULT result = folderif->SetNameOf( NULL, m->get_pidl(index).get(),
        T2COLE( new_name), SHGDN_INFOLDER, &tmp_pidl);
    if (result != S_OK || !tmp_pidl)
        {
        throw favlib_com_error( result);
        }
    auto_pidl new_pidl( tmp_pidl); // make sure tmp_pidl gets deallocated

    // If it is a folder, rename its registry key
    if (is_folder( index))
        {
        auto_memptr<WCHAR> from_subfolder = m->get_relative_path(index);
        const TCHAR *from_subfolder_tchar = W2CT(from_subfolder.get());
        auto_memptr<WCHAR> to_subfolder = m->get_relative_path(T2CW(new_name));
        favs_order::copy( from_subfolder_tchar, W2CT(to_subfolder.get()));
        favs_order::remove( from_subfolder_tchar);
        }

    // Update our favorites array
    m->favs[index] = fav_item( folderif, new_pidl.get());
    m->save_order();
    }


// Removes a child item.
void fav_folder::remove( unsigned index)
    {
    USES_CONVERSION;

    // Delete the file or folder and registry key
    auto_memptr<WCHAR> full_path = m->get_full_path( index);
    if (is_folder( index))
        {
        recursive_unlink( W2CT( full_path.get()));
        favs_order::remove( W2CT( m->get_relative_path( index).get()));
        }
    else
        {
        unlink_file( W2CT( full_path.get()));
        }

    // Remove child item from our favorites array
    m->favs.erase_item( index);
    m->save_order();
    }


// Creates a new internet shortcut, inserting it at index position
void fav_folder::create(
    const TCHAR *name, const TCHAR *url, unsigned index)
    {
    USES_CONVERSION;

    // Build the absolute path of the new item
    auto_memptr<WCHAR> file_name = concat_path( T2CW(name), L".url", false);
    auto_memptr<WCHAR> full_path = m->get_full_path( file_name.get());

    // Check for file name collision, so we don't overwrite anything
    if (path_exists( full_path.get()))
        {
        throw favlib_win_error( ERROR_ALREADY_EXISTS);
        }

    // Create an internet shortcut file
    inet_shortcut_cache shortcut;
    shortcut.set_url( url);
    shortcut.save( full_path.get());

    // Add an entry to our favorites array
    auto_pidl new_pidl = get_relative_pidl( m->get_iface(), file_name.get());
    m->favs.insert_item( index) = fav_item( m->get_iface(), new_pidl.get());
    m->save_order();
    }


// Creates a new folder, inserting it at index position
void fav_folder::mkdir( const TCHAR *name, unsigned index)
    {
    USES_CONVERSION;

    // Create a new folder in the filesystem
    const WCHAR *new_name = T2CW(name);
    auto_memptr<WCHAR> full_path = m->get_full_path( new_name);
    if (CreateDirectory( W2CT(full_path.get()), NULL) == 0)
        {
        throw favlib_win_error();
        }

    // Add an entry to our favorites array
    auto_pidl new_pidl = get_relative_pidl( m->get_iface(), new_name);
    m->favs.insert_item( index) = fav_item( m->get_iface(), new_pidl.get());
    m->save_order();
    }


// Moves an item from one position to another within the same folder
void fav_folder::move( unsigned from_index, unsigned to_index)
    {
    m->favs.move_item( to_index, from_index);
    m->save_order();
    }


// Moves an item from src_folder at src_index to this folder at dest_index
void fav_folder::move(
    fav_folder &from_folder, unsigned from_index, unsigned to_index)
    {
    USES_CONVERSION;

    // If source & destination folders are the same, just change the index
    if (!cstricmp( from_folder.m->subfolder.get(), m->subfolder.get()))
        {
        from_folder.move( from_index, to_index);
        return;
        }

    // Build the absolute paths
    const WCHAR *name = from_folder.m->get_parse_name( from_index);
    auto_memptr<WCHAR> from_path = from_folder.m->get_full_path( name);
    auto_memptr<WCHAR> to_path = m->get_full_path( name);

    // Move the file or folder in the filesystem
    if (path_exists( to_path.get()))
        {
        throw favlib_win_error( ERROR_ALREADY_EXISTS);
        }
    if (MoveFile( W2CT(from_path.get()), W2CT(to_path.get())) == 0)
        {
        throw favlib_win_error();
        }

    // If it is a folder, move its registry key
    if (from_folder.is_folder( from_index))
        {
        auto_memptr<WCHAR> from_subfolder =
            from_folder.m->get_relative_path( from_index);
        const TCHAR *from_subfolder_tchar = W2CT(from_subfolder.get());
        auto_memptr<WCHAR> to_subfolder = m->get_relative_path( name);
        favs_order::copy( from_subfolder_tchar, W2CT( to_subfolder.get()));
        favs_order::remove( from_subfolder_tchar);
        }

    // Add an entry to the destination favorites array
    auto_pidl new_pidl = get_relative_pidl( m->get_iface(), name);
    m->favs.insert_item( to_index) = fav_item( m->get_iface(), new_pidl.get());
    m->save_order();

    // Remove the entry from the source favorites array
    from_folder.m->favs.erase_item( from_index);
    from_folder.m->save_order();
    }


// Returns true if the given file name ends with ".url" (case insensitive).
bool is_doturl_fname( const WCHAR *fname)
    {
    size_t length = cstrlen( fname);
    return length >= 4 && !cstricmp( fname + length - 4, L".url");
    }


// Copies an item from src_folder at src_index to this folder at dest_index
void fav_folder::copy( fav_folder &from_folder, unsigned from_index,
    unsigned to_index, const TCHAR *new_name /* = 0 */ )
    {
    USES_CONVERSION;

    // Build the absolute paths
    const WCHAR *from_name = from_folder.m->get_parse_name( from_index);
    auto_memptr<WCHAR> from_path = from_folder.m->get_full_path( from_name);
    const WCHAR *to_name = from_name;
    auto_memptr<WCHAR> renamed;
    if (new_name)
        {
        if (!from_folder.is_folder( from_index) && is_doturl_fname( from_name))
            {
            renamed = concat_path( T2CW(new_name), L".url", false);
            to_name = renamed.get();
            }
        else
            {
            to_name = T2CW(new_name);
            }
        }
    auto_memptr<WCHAR> to_path = m->get_full_path( to_name);

    // (If source & destination paths are the same, copy functions will fail.)

    // Copy the file or folder in the filesystem
    if (from_folder.is_folder( from_index))
        {
        copy_directory( W2CT(from_path.get()), W2CT( to_path.get()));
        }
    else
        {
        copy_file( W2CT(from_path.get()), W2CT( to_path.get()));
        }

    // If it is a folder, copy its registry key
    if (from_folder.is_folder( from_index))
        {
        auto_memptr<WCHAR> from_subfolder =
            from_folder.m->get_relative_path( from_index);
        auto_memptr<WCHAR> to_subfolder = m->get_relative_path( to_name);
        favs_order::copy(
            W2CT(from_subfolder.get()), W2CT( to_subfolder.get()));
        }

    // Add an entry to the destination favorites array
    auto_pidl new_pidl = get_relative_pidl( m->get_iface(), to_name);
    m->favs.insert_item( to_index) = fav_item( m->get_iface(), new_pidl.get());
    m->save_order();
    }


void fav_folder::sort( less_predicate *pred)
    {
    std::sort( m->favs.begin(), m->favs.end(), pred);
    m->save_order();
    }


bool fav_folder::wait_for_change(
    bool include_subfolders, int milliseconds) const
    {
    // We don't monitor changes to sort order, because that is stored in the
    // Windows registry, and Win95 does not support RegNotifyChangeKeyValue().

    filesystem_change change( m->get_full_path().get(), include_subfolders);
    return change.wait( milliseconds);
    }


/////////////////////////////////////////////////////////////////////////////
// Exception class implementations


void favlib_win_error::format() throw()
    {
#ifdef UNICODE
    TCHAR tmp_message[sizeof(m_message)];
    if (FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, NULL, code, 0,
        tmp_message, sizeof(tmp_message)/sizeof(tmp_message)[0], NULL) != 0)
        {
        if (WideCharToMultiByte( CP_ACP, 0, tmp_message, -1,
            m_message, sizeof(m_message), NULL, NULL) != 0)
            {
            return;
            }
        }
#else // !UNICODE
    if (FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, NULL, code, 0,
        m_message, sizeof(m_message)/sizeof(m_message)[0], NULL) != 0)
        {
        return;
        }
#endif // UNICODE
    strcpy( m_message, "Windows function failed");
    }


/////////////////////////////////////////////////////////////////////////////
// EOF
