/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is favlib, the Windows-compatible bookmarks library.
 *
 * The Initial Developer of the Original Code is
 * Forest Wilkinson.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Forest Wilkinson
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


/*
favlib.h

Copyright 2004-2008 Forest Wilkinson
*/


#include <tchar.h>
#include <windows.h>
#include <exception>
#include <memory>


/////////////////////////////////////////////////////////////////////////////
// Exception classes


class favlib_exception: public std::exception
    {
    public:
        virtual const char *what() const throw()
            { return "an exception was thrown from the favlib library"; }
    protected:
        favlib_exception() {} // don't use directly; use a derived class.
    };


// Indicates a Win32 API error, and carries an error code from winerror.h.
class favlib_win_error: public favlib_exception
    {
    public:
        favlib_win_error() throw() { code = GetLastError(); format(); }
        explicit favlib_win_error( DWORD e) throw(): code(e) { format(); }
        virtual const char *what() const throw() { return m_message; }
        DWORD code;
    private:
        void format() throw(); // puts a windows error string in m_message
        char m_message[160];
    };


class favlib_com_error: public favlib_win_error
    {
    public:
        explicit favlib_com_error( HRESULT e = S_OK) throw()
            : favlib_win_error(e) {}
    };


class favlib_bad_alloc: public favlib_com_error
    {
    public:
        favlib_bad_alloc() throw(): favlib_com_error( E_OUTOFMEMORY) {}
    };


/////////////////////////////////////////////////////////////////////////////
// fav_folder - the public interfaces for manipulating favorites


// WARNING: You must call CoInitialize[Ex]() before using this library.


// Not for STL containers; copy & assignment steal ownership of internal data.
class fav_folder
    {
    public:
        explicit fav_folder( bool with_files = true);
        fav_folder( const fav_folder &parent, unsigned child_index,
            bool with_files = true); // initialize to a subfolder
        ~fav_folder();

        // these get information about the current folder (*this):
        bool has_subfolder() const;
        unsigned length() const;

        // these get information about a child item: (without range checking)
        bool is_folder( unsigned index) const;
        bool has_subfolder( unsigned index) const;
        const TCHAR *get_name( unsigned index) const;
        const TCHAR *get_url( unsigned index);

        // these alter a child item:
        // (only index args with default value do range check; others do not.)
        void set_url( unsigned index, const TCHAR *url);
        void rename( unsigned index, const TCHAR *new_name);
        void remove( unsigned index);
        void create( const TCHAR *name, const TCHAR *url, unsigned index = -1);
        void mkdir( const TCHAR *name, unsigned index = -1);
        void move( unsigned from_index, unsigned to_index);
        void move( fav_folder &from_folder, unsigned from_index,
            unsigned to_index = -1);
        void copy( fav_folder &from_folder, unsigned from_index,
            unsigned to_index = -1, const TCHAR *new_name = 0);

        // interface to sortable items & callback function type for sort():
        class sortable_item
            {
            public:
                virtual const TCHAR *get_name() const = 0;
                virtual bool is_folder() const = 0;
                virtual bool has_subfolder() const = 0;
            };
        // prototype of comparison function called by sort()
        // (implementations should return true if x < y)
        typedef bool less_predicate(
            const sortable_item &x, const sortable_item &y);
        virtual void sort( less_predicate *pred);

        // Returns true if a change occurred, or false if the timeout elapsed.
        bool wait_for_change(
            bool include_subfolders, int milliseconds = -1) const;

    private:
        struct private_data;
        friend struct private_data;
        std::auto_ptr<private_data> m;
    };


inline std::auto_ptr<fav_folder> get_fav_root( bool with_files = true)
    {
    return new fav_folder( with_files);
    }


inline std::auto_ptr<fav_folder> get_fav_subfolder(
    const fav_folder &parent, unsigned child_index, bool with_files = true)
    {
    return new fav_folder( parent, child_index, with_files);
    }


/////////////////////////////////////////////////////////////////////////////
// EOF
