/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is favlib, the Windows-compatible bookmarks library.
 *
 * The Initial Developer of the Original Code is
 * Forest Wilkinson.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Forest Wilkinson
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


/////////////////////////////////////////////////////////////////////////////
// testing module


#include "favlib.h"

#include <iostream>
#include <cerrno>
#include <climits>
#include <cstdlib>
#include <cstring>

using namespace std;

#ifdef UNICODE
    #define tcout std::wcout
#else
    #define tcout std::cout
#endif // UNICODE


// Like atoi(), but raises exception on failure
unsigned atou( const char *str)
    {
    errno = 0;
    char *stopper;
    unsigned long val = strtoul( str, &stopper, 10);
    if (*stopper || errno || val > UINT_MAX)
        {
        cerr << "bad numeric value: " << str << endl;
        }
    return val;
    }


struct favidx
    {
    favidx(): index(-1) {}
    favidx( auto_ptr<fav_folder> &f, unsigned i): folder(f), index(i) {}
    favidx( favidx &other): folder(other.folder), index(other.index) {}
    explicit favidx( char *path, bool index_last = false);
    favidx& operator=( favidx &other)
        {folder = other.folder; index = other.index;}

    auto_ptr<fav_folder> folder;
    unsigned index;
    };

/*
Parses the given path as relative to the root favorites folder.

If path refers to a folder and index_last is true,
or if path refers to a non-folder,
stores the parent fav_folder object in this->folder & leaf index in this->index.

If path refers to a folder and index_last is false,
stores the fav_folder object in this->folder and -1 in this->index.
*/
favidx::favidx( char *path, bool index_last)
    {
    folder = get_fav_root();
    index = -1;

    auto_ptr<fav_folder> prev_folder;
    unsigned prev_index;
    char *next;

    while ((next = strtok( path, "/\\")) != 0)
        {
        path = NULL;
        index = atou( next);

        if (!folder->is_folder( index))
            {
            return;
            }

        prev_folder = folder;
        prev_index = index;
        folder = get_fav_subfolder( *folder, index);
        index = -1;
        }

    if (index_last && prev_folder.get())
        {
        folder = prev_folder;
        index = prev_index;
        }
    }


void do_show( char **arguments)
    {
    favidx target( arguments[0]);

    if (target.index == -1)
        {
        tcout << "(has " << (target.folder->has_subfolder() ? "a" : "no")
            << " subfolder)" << endl;

        for (unsigned n = 0; n != target.folder->length(); ++n)
            {
            tcout << n << (target.folder->is_folder(n) ? _T("(f)") : _T(""))
                << _T(": ") << target.folder->get_name(n) << endl;
            }

        }
    else
        {
        tcout << target.folder->get_url(target.index) << endl;
        }
    }


void do_wait( char **arguments)
    {
    favidx target( arguments[0]);

    tcout << "waiting for changes..." << endl;
    for(;;)
        {
        if (target.folder->wait_for_change( true))
            {
            tcout << "change notification received" << endl;
            }
        }
    }


void do_seturl( char **arguments)
    {
    favidx target( arguments[0], true);
    if (target.index == -1 || target.folder->is_folder(target.index))
        {
        cerr << "you can't set the url of a folder!" << endl;
        exit(1);
        }
    const char *url = arguments[1];

    target.folder->set_url( target.index, url);
    }


void do_hassub( char **arguments)
    {
    favidx target( arguments[0], true);
    if (target.index == -1 || !target.folder->is_folder(target.index))
        {
        cerr << "hassub operates on subfolders" << endl;
        exit(1);
        }

    bool hassub = target.folder->has_subfolder( target.index);
    tcout << "has " << (hassub ? "a" : "no") << " subfolder" << endl;
    }


void do_rename( char **arguments)
    {
    favidx target( arguments[0], true);
    if (target.index == -1)
        {
        cerr << "you can't rename the root favorites folder!" << endl;
        exit(1);
        }
    const char *newname = arguments[1];

    target.folder->rename( target.index, newname);
    }


void do_remove( char **arguments)
    {
    favidx target( arguments[0], true);
    if (target.index == -1)
        {
        cerr << "you can't remove the root favorites folder!" << endl;
        exit(1);
        }

    target.folder->remove( target.index);
    }


void do_create( char **arguments)
    {
    favidx parent( arguments[0], false);
    const char *name = arguments[1];
    const char *url = arguments[2];
    parent.folder->create( name, url);
    }


void do_mkdir( char **arguments)
    {
    favidx parent( arguments[0], false);
    const char *name = arguments[1];
    parent.folder->mkdir( name);
    }


void do_move( char **arguments)
    {
    favidx source( arguments[0], true);
    if (arguments[1][0] == '/')
        {
        favidx dest( arguments[1], false);
        dest.folder->move( *source.folder, source.index);
        }
    else
        {
        unsigned dest_index = atou( arguments[1]);
        source.folder->move( source.index, dest_index);
        }
    }


void do_copy( char **arguments)
    {
    favidx source( arguments[0], true);
    favidx target( arguments[1], false);
    target.folder->copy( *source.folder, source.index);
    }


bool is_less(
    const fav_folder::sortable_item &x, const fav_folder::sortable_item &y)
    {
    return _tcsicmp( x.get_name(), y.get_name()) < 0;
    }


void do_sort( char **arguments)
    {
    favidx target( arguments[0], false);
    target.folder->sort( is_less);
    }


void usage( const char *prog)
    {
    const char *p = strrchr( prog, '\\');
    if (p)
        {
        prog = p + 1;
        }
    cerr << "usage: " << prog << " <command> [<arg1> [<arg2> [...]]]]" << endl;
    cerr << "commands:\n";
    cerr << " show\t<path>\n";
    cerr << " hassub\t<path>\n";
    cerr << " wait\t<path>" << endl;
    cerr << " seturl\t<path> <url>\n";
    cerr << " rename\t<path> <newname>\n";
    cerr << " remove\t<path>\n";
    cerr << " create\t<path> <name> <url>\n";
    cerr << " mkdir\t<path> <name>\n";
    cerr << " move\t<source-path> <target-index>|<target-path>\n";
    cerr << " copy\t<source-path> <target-path>\n";
    cerr << " sort\t<path>\n";
    cerr << "\t<path> is a list of slash-separated indexes, e.g. \"/0/17\"\n";
    }


typedef struct
    {
    const char *name;
    void (*func)( char **arguments);
    unsigned nargs; // number of args
    } command_info;
const command_info g_commands[] =
    {
        { "show",   do_show,    1},
        { "hassub", do_hassub,  1},
        { "wait",   do_wait,    1},
        { "seturl", do_seturl,  2},
        { "rename", do_rename,  2},
        { "remove", do_remove,  1},
        { "create", do_create,  3},
        { "mkdir",  do_mkdir,   2},
        { "move",   do_move,    2},
        { "copy",   do_copy,    2},
        { "sort",   do_sort,    1},
        { 0, 0, 0}
    };


int main( int argc, char **argv)
    {
    CoInitialize( NULL);

    // check usage
    if (argc < 2 || argv[1][0] == '-')
        {
        usage( argv[0]);
        exit(1);
        }
    char **arguments = argv + 1;
    unsigned nargs = argc - 2;

    // run command
    try
        {
        const char *cmdname = *arguments++;
        for (const command_info *cmd = g_commands; cmd->name; ++cmd)
            {
            if (!strcmp( cmdname, cmd->name))
                {
                if (nargs != cmd->nargs)
                    {
                    cerr << cmdname <<
                        " requires " << cmd->nargs << " arguments" << endl;
                    exit(1);
                    }
                cmd->func( arguments);
                exit(0);
                }
            }
        usage( argv[0]);
        exit(1);
        }
    catch (std::exception &e)
        {
        cerr << "exception caught: " << e.what() << endl;
        }
    return 1;
    }


/////////////////////////////////////////////////////////////////////////////
// EOF
